// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package mock_test

import (
	"bytes"
	"go/ast"
	"go/format"
	"go/parser"
	"go/token"

	"git.sr.ht/~nelsam/hel/dep"
	"github.com/poy/onpar/expect"
)

const packagePrefix = "package foo\n\n"

func source(expect expect.Expectation, pkg string, decls []ast.Decl, scope *ast.Scope) string {
	var buf bytes.Buffer
	f := &ast.File{
		Name:  &ast.Ident{Name: pkg},
		Decls: decls,
		Scope: scope,
	}
	err := format.Node(&buf, token.NewFileSet(), f)
	expect(err).To(not(haveOccurred()))
	return buf.String()
}

func parse(expect expect.Expectation, code string) *ast.File {
	f, err := parser.ParseFile(token.NewFileSet(), "", packagePrefix+code, 0)
	expect(err).To(not(haveOccurred()))
	expect(f).To(not(beNil()))
	return f
}

func depSpec(expect expect.Expectation, spec string, opts ...dep.Opt) dep.Dependency {
	s := typeSpec(expect, spec)
	return dep.New(s.Name, s.Type, opts...)
}

func typeSpec(expect expect.Expectation, spec string) *ast.TypeSpec {
	f := parse(expect, spec)
	expect(f.Scope.Objects).To(haveLen(1))
	for _, obj := range f.Scope.Objects {
		spec, ok := obj.Decl.(*ast.TypeSpec)
		expect(ok).To(beTrue())
		if spec.TypeParams == nil {
			// For ease of testing, we just want to ensure that this field is
			// non-nil.
			spec.TypeParams = &ast.FieldList{}
		}
		return spec
	}
	return nil
}

func function(expect expect.Expectation, decl string) *ast.FuncDecl {
	f := parse(expect, decl)
	expect(f.Scope.Objects).To(haveLen(1))
	for _, obj := range f.Scope.Objects {
		decl, ok := obj.Decl.(*ast.FuncDecl)
		expect(ok).To(beTrue())
		return decl
	}
	return nil
}

func method(expect expect.Expectation, spec *ast.TypeSpec, names ...string) *ast.FuncType {
	inter, ok := spec.Type.(*ast.InterfaceType)
	expect(ok).To(beTrue())
	l := inter.Methods.List
	if len(names) > 0 {
		l = nil
		for _, mn := range inter.Methods.List {
			for _, n := range names {
				if len(mn.Names) == 0 {
					continue
				}
				if mn.Names[len(mn.Names)-1].Name != n {
					continue
				}
				l = append(l, mn)
				break
			}
		}
	}
	expect(l).To(haveLen(1))
	f, ok := l[0].Type.(*ast.FuncType)
	expect(ok).To(beTrue())
	return f
}
